/************************************************************************/
/* Z_trans.c - This program finds the roots (zeros) associated with a	*/
/*  polynomial.  For example, the equation c1 * x^2 + c2 * x + c3 = 0.	*/
/*  It can be represented by the vector:  {c1, c2, c3}			*/
/*  If the vector file is supplied to Z_trans, it will solve for the	*/
/*  values of "x" where the above equation is true (is zero).		*/
/*  This operation is also called, the Z-transform in discrete systems.	*/
/*                                                                      */
/* Operation: This program works by first getting a quick initial 	*/
/* estimate of the roots by performing the long-division of the 	*/
/* polynomial.  It then iterates a genetic algorithm to refine the	*/
/* answer by zeroing-in on the roots.					*/
/*                                                                      */
/* Options:                                                             */
/*      -max xx    - Constrains the search area by setting a maximum	*/
/*			practical value for the magnitude of "x".	*/
/*		     By default, the max-value is set to 100.0.		*/
/*      -min xx    - Constrains the search area by setting a minimum	*/
/*			practical value for the magnitude of "x".	*/
/*		     By default, the max-value is set to 100.0.		*/
/*	-eps  xx   - Sets the maximum error value.  Search is stopped	*/
/*			when error gets smaller than this value.	*/
/*		     By default, epsilon is 0.000001.			*/
/*	-v	   - Verbose.  Provides status while running.		*/
/*                                                                      */
/* Example Usage:                                                       */
/*      z_trans infile  						*/
/*                                                                      */
/* File Types:                                                          */
/*      In:   Complex or Real Vector					*/
/*      Out:  Complex or Real Vector (of roots)				*/
/*                                                                      */
/* From:  www.xgraph.org/numutil					*/
/*        xgraphadmin@xgraph.org					*/
/*                                                                      */
/* Copyright (C) 2002 CHein                                             */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, Version 2 of the GPL.                  */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/* You should have received a copy of the GNU General Public License    */
/* along with this program; if not, write to the Free Software          */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307. */
/************************************************************************/

#include "commonlibs.c"

double	maxx=100.0, minx=-100.0;
double	eps = 0.000001,
	grid_step = 0.01;

int verbose = 0;
int  n_elements;
int seed=123;


double Random()	/* Return random value between 0.0 and 1.0. */
{
 #ifdef __MINGW32__
  static double RCONST = (double)1.0 / (double)32767.0;
  return (double) rand() * RCONST;
 #else
  static double RCONST = (double)1.0 / (double)2147483647.0; /* set up the rand normalizing constant */
  return (double)random() * RCONST;
 #endif
}



/* FACTOR - Perform the Long Division of two poly-nomials. */
void FACTOR( complex ctmp, complex *w, int *order )
{
 int i, j, k;
 complex *quotient, *divisor, *subtrahend, *dividend;

 quotient = (complex *)malloc( *order * sizeof(complex) );
 divisor = (complex *)malloc( *order * sizeof(complex) );
 subtrahend = (complex *)malloc( *order * sizeof(complex) );
 dividend = (complex *)malloc( *order * sizeof(complex) );

 divisor[1] = COMP(1,0);
 divisor[0] = ctmp;
 for (i=0; i < *order; i++)  dividend[i] = w[i];

 for (i=0; i!=*order-1; i++) 
  { /*1*/
   quotient[i] = dividend[i];
   for (j=0; j!=2; j++) 
    subtrahend[j] = CMULT( divisor[j], quotient[i] );
   for (j=0; j!=2; j++) 
    dividend[i-j+1] = CADD( dividend[i-j+1], subtrahend[j] );
  } /*1*/
 for (i = 0; i!=*order; i++) w[i] = quotient[i];
 *order = *order - 1;

 free( quotient );
 free( divisor );
 free( subtrahend );
 free( dividend );
} /*Factor*/




/*Evaluate Function*/
double func( double x, double y, complex *w, int n_elements )
{
 complex zi, ctmp, sum;
 int n;
 double rtmp;

 /*printfln(x,' ',y);*/
 ctmp = COMP(x, y);
 zi = CDIV(COMP(1.0,0.0), ctmp);
 sum = COMP(0.0,0.0);
 for (n = 0; n  < n_elements; n++) 
  { /*c*/
   ctmp = CMULT( ctmp, zi);
   // printf("		ctmp=%f %f w[%d] =%f %f\n", ctmp.r, ctmp.i, n, w[n].r, w[n].i ); 
   sum = CADD( sum, CMULT( ctmp, w[n]));
  } /*c*/      

 // printf("		%d f(%f, %f) = %f %f\n", n_elements, x, y, sum.r, sum.i); 
 
 if ((fabs(sum.r) < 10000.0) && (fabs(sum.i) < 10000.0))  /* Prevent overflow. */
  rtmp = CMAG(sum); else rtmp = 100000.0;

  return rtmp;
} /*Evaluate f*/







complex FIND_ZERO( complex *w, int nel )
{
 double x=1, y=1, dx, dy, bestval, curval, newval;
 int swtch=0, count;

 if (verbose) printf("Finding next root %d.\n", nel);
 dx = 0.1 * grid_step;
 dy = 0.1 * grid_step; 
 do
  { /*loop*/
   count = 0; 
   curval = func( x, y, w, nel );
   bestval = bestval;
   while ( (curval > eps) && (count < 2000) && (fabs(x) + fabs(dx) < maxx) 
		&& (fabs(y) + fabs(dy) < maxx) && (fabs(x) + fabs(dx) > minx) 
		&& (fabs(y) + fabs(dy) > minx)) 
    { /*2*/
     count = count + 1;
	/*printfln('	func(',x,',',y,') =',func(x,y,w));*/
	/* printf("%f\n", f(x,y,w)); */

     newval = func(x + dx, y, w, nel);
     if (newval > curval)
      { /*worse*/
	dx = -0.25 * dx;
	// printf("	worse dx = %g\n", dx );
      } /*worse*/
     else
      { /*better*/
	x = x + dx;
	dx = 1.5 * dx;
	curval = newval;
	// printf("	Better %g, x=%g  dx = %g\n", curval, x, dx );
      } /*better*/

     newval = func(x, y + dy, w, nel);
     if (newval > curval)
      { /*worse*/
	dy = -0.25 * dy;
	// printf("	worse dy = %g\n", dy );
      } /*worse*/
     else
      { /*better*/
	y = y + dy;
	dy = 1.5 * dy;
	curval = newval;
	// printf("	Better %g, y=%g  dy = %g\n", curval, y, dy );
      } /*better*/

    } /*2*/
   // printf("	Iter: func(%g, %g) = %g	(count=%d)\n", x, y, func(x,y, w,nel), count );

   if (curval >= 100.0 * eps)
    {
     swtch = swtch + 1;
     dx = 0.1 * grid_step;  dy = 0.1 * grid_step; 
     if (x > 0.0)  x = -1.0;  else  x = 1.0;
     if (y > 0.0)  y = -1.0;  else  y = 1.0;
     if (swtch > 4)
      {
       do
	{ x = Random() * 3.0 - 0.75; }
       while (fabs(x) < 0.2);
       do
	{ y = Random() * 3.0 - 0.75; }
       while (fabs(y) < 0.2);
        swtch = 0;
      }
     else
      switch ((int)((x*2.0+y)+0.5))	/* Initial sweep in cardinal directions. */
       {
        case 3: x= -1.0; break;
        case 1: y = 1.0; break;
        case -1: y = -1.0; break;
        case -3: x = 1.0; break;
       }
    }
  } /*loop*/
 while (func(x,y, w,nel) > 100.0 * eps);

 if (verbose) printf("%d  Found Z: F(%f, %f) = %f	(refinement count=%d)\n", nel, x, y, func(x,y,w,nel), count );
 if ((count>=100) || (fabs(x) >= maxx) || (fabs(y) >= maxx)) 
   { x=0.0; y=0.0; printf("	restarting search (%f, %f)\n", x, y); }

 return COMP(x,y);

} /*function FIND_ZERO*/




/* ZTRANS - Solves and plots Z-transform (in Z-plane). */
void ZTRANS( int n_elements, complex *coef, complex *zlist )
{
 int j, ntodo, k=0;
 complex ctmp;
 FILE *zplot;

 zplot = fopen("zplot.dat","w"); 

 ntodo = n_elements;
 while (ntodo > 1) 
 { /*a*/
   ctmp = FIND_ZERO( coef, ntodo ); 

   if (ntodo > 2)       /* Factor next zero from polynomial. */
    FACTOR( ctmp, coef, &ntodo );
   else
    ntodo = ntodo - 1;
   fprintf(zplot,"%g	%g\n", ctmp.r, ctmp.i);

   zlist[k++] = ctmp;
 } /*a*/
} /*ZTRANS*/





int main( int argc, char *argv[] )
{
 int arg=1, j, k=0, X;
 struct file_kind *infile=0, *outfile=0;
 char fname1[999], fname2[999]="", line[500], word[100], trsh[100];
 complex *coef, *zlist;
 double x, y;

 while (argc > arg) 
  {
   if (argv[arg][0] == '-')
    {
     if (strcmp(argv[arg],"-v") == 0) { verbose = 1;  printf("Setting verbose mode,\n"); }
     else
     if (strcmp(argv[arg],"-max") == 0)
      {
	arg++;
	if ((arg == argc) || (sscanf(argv[arg],"%lf",&maxx) != 1))
	 { printf("Error ztrans: Bad value after %s.\n", argv[arg-1]);  exit(1); }
      }
     else
     if (strcmp(argv[arg],"-min") == 0)
      {
	arg++;
	if ((arg == argc) || (sscanf(argv[arg],"%lf",&minx) != 1))
	 { printf("Error ztrans: Bad value after %s.\n", argv[arg-1]);  exit(1); }
      }
     else
     if (strcmp(argv[arg],"-eps") == 0)
      {
	arg++;
	if ((arg == argc) || (sscanf(argv[arg],"%lf",&eps) != 1))
	 { printf("Error ztrans: Bad value after %s.\n", argv[arg-1]);  exit(1); }
      }
     else printf("Unknown option '%s'\n", argv[arg] );
    }
   else
    { /*file_name*/
     if (k==0)  
      { 
       strcpy(fname1,argv[arg]);
       infile = file_open_read( fname1 );
       k = k + 1;
      }
     else
     if (k==1)  
      {
       strcpy(fname2,argv[arg]);
       k = k + 1;
      }
     else
     {printf("Error ztrans: Too many file-names on command line.\n"); exit(1);}
   } /*file_name*/

  arg++; 
 }

 if (k < 1) { printf("ERROR ztrans: Missing input file on command line.\n"); exit(1); }
 if (infile->fpt == 0) {printf("ERROR ztrans: Can't open input file '%s'.\n",fname1); exit(1);}
 if (infile->dim2 > 0)
  {printf("ERROR ztrans: Can't solve multi-dimensional vectors.\n");  exit(1);}
 if (k < 2)  /* No output file specified, so make default name by appending suffix. */
  { 
    strcpy(fname2,fname1);
    j = strlen(fname2)-1;  while ((j >= 0) && (fname2[j]!='.')) j--;
    if (j < 0) j = strlen(fname2);
    fname2[j] = '\0';
    strcat(fname2,".new");
  }
 if (strcmp(fname1,fname2) == 0) {printf("ERROR ztrans: Can't over-write source.\n"); exit(1);}

 printf("       (Writing output to %s)\n", fname2 );
 outfile = file_open_write( fname2, infile->kind, infile->dim1, infile->dim2 );
 if (outfile->fpt==0) {printf("ERROR ztrans: Can't open output file '%s'.\n",fname2); exit(1);}

 /* Initialize random seed for refinement searches. */
 #ifdef __MINGW32__
  srand(seed);
 #else
  srandom(seed);
 #endif

 /* First determine the number of items in input file. */
 n_elements = 0;
 Read_Line( infile->fpt, line, 500 );
 while (! feof( infile->fpt ))
  {
   Next_Word(line,trsh,word," \t,()");
   if (word[0] != '\0')
    {
     n_elements++;
    }
   Read_Line( infile->fpt, line, 500 );
  }
 fclose(infile->fpt);
 free( infile );
 
 coef = (complex *)malloc( (n_elements + 1) * sizeof(complex) );
 zlist = (complex *)malloc( (n_elements + 1) * sizeof(complex) );

 j = 0;
 infile = file_open_read( fname1 );
 Read_Line( infile->fpt, line, 500 );
 while (! feof( infile->fpt ))
  {
   if (j >= n_elements) { printf("ERROR ztrans: too many items in input file (>%d.\n", n_elements);  exit(1); }
   Next_Word(line,trsh,word," \t,()");
   if (word[0] != '\0')
    {

     if (infile->kind == 'R')
      {
       if (sscanf(word,"%d",&X)!=1) printf("ERROR ztrans: Reading index '%s' on line %d.\n",word,j+1);
       Next_Word(line,trsh,word," \t,()");
       if (sscanf(word,"%lf",&y)!=1) printf("ERROR ztrans: Reading value '%s' on line %d.\n",word,j+1);
       coef[j] = COMP( y, 0.0 );
      }
     else
      { /* infile->kind=='C' */
       if (sscanf(word,"%lf",&x)!=1) printf("ERROR ztrans: Reading real '%s' on line %d.\n",word,j+1);
       Next_Word(line,trsh,word," \t,()");
       if (sscanf(word,"%lf",&y)!=1) printf("ERROR ztrans: Reading imaginary '%s' on line %d.\n",word,j+1);
       coef[j] = COMP( x, y );
      }
     j = j + 1;
    }
   Read_Line( infile->fpt, line, 500 );
  }
 fclose(infile->fpt);

 ZTRANS( n_elements, coef, zlist );

 for (j=0; j < n_elements - 1; j++)
  {
   if (outfile->kind == 'R')
    fprintf(outfile->fpt,"%d	%g\n", j, zlist[j].r );
   else
    fprintf(outfile->fpt,"	%g	%g\n", zlist[j].r, zlist[j].i );
  }
 fclose(outfile->fpt);
 
} /*main*/
